/*
 * livepatch_bsc1106191
 *
 * Fix for CVE-2018-10938, bsc#1106191
 *
 *  Upstream commit:
 *  40413955ee26 ("Cipso: cipso_v4_optptr enter infinite loop")
 *
 *  SLE12(-SP1) commit:
 *  not affected
 *
 *  SLE12-SP2 and -SP3 commit:
 *  dfcceea9d5f7cb02078e0e0176279d50a061eb35
 *
 *  SLE15 commit:
 *  4af59ddef75512b3e8bcaa0cb3bc06b56001cb50
 *
 *
 *  Copyright (c) 2018 SUSE
 *  Author: Nicolai Stange <nstange@suse.de>
 *
 *  Based on the original Linux kernel code. Other copyrights apply.
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, see <http://www.gnu.org/licenses/>.
 */

#include <linux/kernel.h>
#include <linux/module.h>
#include <linux/skbuff.h>
#include <linux/ip.h>
#include "livepatch_bsc1106191.h"

/* patched */
unsigned char *klp_cipso_v4_optptr(const struct sk_buff *skb)
{
	const struct iphdr *iph = ip_hdr(skb);
	unsigned char *optptr = (unsigned char *)&(ip_hdr(skb)[1]);
	int optlen;
	int taglen;

	for (optlen = iph->ihl*4 - sizeof(struct iphdr); optlen > 0; ) {
		/*
		 * Fix CVE-2018-10938
		 *  -3 lines, +11 lines
		 */
		switch (optptr[0]) {
		case IPOPT_CIPSO:
			return optptr;
		case IPOPT_END:
			return NULL;
		case IPOPT_NOOP:
			taglen = 1;
			break;
		default:
			taglen = optptr[1];
		}
		optlen -= taglen;
		optptr += taglen;
	}

	return NULL;
}
